package org.kantega.jexmec.guice;

/*
 * Copyright 2009 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import com.google.inject.*;
import com.google.inject.binder.AnnotatedBindingBuilder;
import com.google.inject.binder.ScopedBindingBuilder;
import com.google.inject.binder.LinkedBindingBuilder;
import com.google.inject.name.Names;
import org.kantega.jexmec.*;

import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.TypeVariable;
import java.util.ArrayList;
import java.util.List;

/**
 * Implementation of {@link PluginLoader} based on <a href="http://code.google.com/p/google-guice/">Google Guice</a>.
 * Plugin classes are located by reading the file {@code META-INF/services/&lt;pluginClassName&gt;/guice.txt} from the
 * ClassLoader. The plugins are then configured in a Guice Module together with all services provided
 * by the ServiceLocator. Finally, plugins are created by using Guice's Injector.findBindingsByType() method.
 */
public class GuicePluginLoader<P extends Plugin>  implements PluginLoader<P> {

    private final ClassLocator<Class<P>> classLocator;

    public GuicePluginLoader(ClassLocator<Class<P>> classLocator) {
        this.classLocator = classLocator;
    }

    public GuicePluginLoader() {
        this.classLocator = null;
    }

    /**
     * Locates, loads and returns any Guice plugins found in the provided ClassLoader
     * @param classLoader the ClassLoader used to locate Guice
     * @param serviceLocator the {@link org.kantega.jexmec.ServiceLocator} where services can be fetched.
     * @return the list of Guice Plugins
     */
    public List<P> loadPlugins(final Class<P> pluginClass, ClassLoader classLoader, final ServiceLocator serviceLocator) {


        ClassLocator<Class<P>> classLocator = this.classLocator != null ? this.classLocator : new DefaultClassLocator<Class<P>>(pluginClass, "guice.txt");
        final List<Class<P>> pluginClasses = classLocator.locateClasses(classLoader);


        ParameterizedType type = new ParameterizedType() {
            public Type[] getActualTypeArguments() {
                return GuicePluginLoader.this.getClass().getTypeParameters()[0].getBounds();
            }

            public Type getRawType() {
                return Plugin.class;
            }

            public Type getOwnerType() {
                return null;
            }
        };


        @SuppressWarnings("unchecked")
        final TypeLiteral<P> typeLiteral = (TypeLiteral<P>) TypeLiteral.get(type);


        Module pluginsModule = new Module() {
            public void configure(Binder binder) {
                for(Class<P> pluginClass : pluginClasses) {
                    if(Plugin.class.isAssignableFrom(pluginClass)) {
                        binder.bind(typeLiteral).annotatedWith(Names.named(pluginClass.getName())).to(pluginClass);
                    } else if(Module.class.isAssignableFrom(pluginClass)) {
                        try {
                            binder.install((Module) pluginClass.newInstance());
                        } catch (InstantiationException e) {
                            throw new RuntimeException(e);
                        } catch (IllegalAccessException e) {
                            throw new RuntimeException(e);
                        }
                    }

                }
            }
        };

        List<P> plugins = new ArrayList<P>();


        List<Module> allModules = new ArrayList<Module>();
        allModules.add(new Module() {
            public void configure(Binder binder) {

                for(final ServiceKey serviceKey : serviceLocator.getServiceKeys()) {


                    Provider serviceProvider = new Provider() {
                        public Object get() {
                            @SuppressWarnings("unchecked")
                            Object service = serviceLocator.lookupService(serviceKey.getServiceClass(), serviceKey.getServiceName());
                            return service;
                        }
                    };

                    // Bind with type literal if type parameter is bounded with Plugin
                    if(serviceKey.getServiceClass().getTypeParameters().length  == 1) {
                        final TypeVariable typeVariable = serviceKey.getServiceClass().getTypeParameters()[0];
                        if(typeVariable.getBounds().length == 1 && typeVariable.getBounds()[0] == Plugin.class) {

                            final TypeLiteral<?> serviceTypeLiteral = TypeLiteral.get(new ParameterizedType() {
                                public Type[] getActualTypeArguments() {
                                    return new Type[]{pluginClass};
                                }

                                public Type getRawType() {
                                    return serviceKey.getClass();
                                }

                                public Type getOwnerType() {
                                    return null;
                                }
                            });


                            AnnotatedBindingBuilder<?> builder = binder.bind(serviceTypeLiteral);
                            @SuppressWarnings("unchecked")
                            ScopedBindingBuilder builder1 = builder.toProvider(serviceProvider);
                            builder1.in(Singleton.class);
                        }
                    }
                    // Bind the class serviceKey
                    @SuppressWarnings("unchecked")
                    Key<Object> key = serviceKey.getServiceName() != null
                                        ? Key.<Object>get(serviceKey.getServiceClass(), Names.named(serviceKey.getServiceName()))
                                        : Key.<Object>get(serviceKey.getServiceClass());

                    LinkedBindingBuilder<?> builder = binder.bind(key);
                    @SuppressWarnings("unchecked")
                    ScopedBindingBuilder builder1 = builder.toProvider(serviceProvider);
                    builder1.in(Singleton.class);

                }
            }


        });
        allModules.add(pluginsModule);

        final Injector injector = Guice.createInjector(allModules);
        

        for(Binding<P> binding : injector.findBindingsByType(typeLiteral)) {

            plugins.add(binding.getProvider().get());
        }
        return plugins;
    }

    public void unloadPlugins(ClassLoader classLoader) {
    }

    public void start() {

    }

    public void stop() {
        
    }
}
