package org.kantega.jexmec.guice;

/*
 * Copyright 2009 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.kantega.jexmec.ServiceLocator;
import org.kantega.jexmec.ServiceKey;

import java.util.*;

import com.google.inject.Guice;
import com.google.inject.Injector;
import com.google.inject.Module;
import com.google.inject.Key;
import com.google.inject.name.Named;
import com.google.inject.name.Names;

/**
 * {@link ServiceLocator} implementation where services are looked up in a Guice module.
 */
public class GuiceServiceLocator implements ServiceLocator {
	private final Map<ServiceKey, Object> services;

    /**
     * Creates a ServiceLocator that will look up service classes in the provided Guice module.
     * @param module the module to look in
     * @param serviceClasses the service classes that will be looked up
     */
    public GuiceServiceLocator(Module module, Set<ServiceKey> serviceClasses) {
        Map<ServiceKey, Object> services = Collections.synchronizedMap(new HashMap<ServiceKey, Object>());
        
        Injector injector = Guice.createInjector(module);

        for (ServiceKey serviceKey : serviceClasses) {

            @SuppressWarnings("unchecked")
            Key<Object> key = serviceKey.getServiceName() != null
                    ? Key.<Object>get(serviceKey.getServiceClass(), Names.named(serviceKey.getServiceName()))
                    : Key.<Object>get(serviceKey.getServiceClass());

        	Object instance =  injector.getInstance(key);
            if (instance == null) {
                if(serviceKey.getServiceName() == null) {
                    throw new IllegalStateException("Guice module " + module.toString() + " contains no bean of type " + serviceKey);
                } else {
                    throw new IllegalStateException("Guice module " + module.toString() + " contains no bean of type " + serviceKey +" named " + serviceKey.getServiceName());
                }
            }
            services.put(serviceKey, instance);
        }
        this.services = Collections.unmodifiableMap(services);
    }

    public Set<ServiceKey> getServiceKeys() {
        return services.keySet();
    }

    public <T> T lookupService(Class<T> serviceClass) {
        return lookupService(serviceClass, null);  
    }

    public <T> T lookupService(Class<T> serviceClass, String serviceName) {
        return serviceClass.cast(services.get(new ServiceKey(serviceClass, serviceName)));
    }
}
